"""
Manejo de error generales
"""

AUTORIZACION_BASE_DE_DATOS = """
    Es necesario obtener autorizacion para el el uso y lectura de base de datos
"""

TRUORA_API_KEY = """
    Debe obtener un API key de truora para consulta bases de datos externas, mas info en: https://dev.truora.com/checks/account/
"""

AUTHORIZATION_EXTERNAL_DB = """
    Para leer fuentes externas, se debe obtener previamente autorizacion del Titular de la información 
"""


class AutorizacionDatosPersonales(Exception):
    """
    Error por falta de autorizacion para uso de datos personales
    """

    def __init__(self):
        super().__init__(AUTORIZACION_BASE_DE_DATOS)


class TruoraApiKeyRequired(Exception):
    """
    Error por falta de api key de truora
    """

    def __init__(self):
        super().__init__(TRUORA_API_KEY)


class TruoraGeneralError(Exception):
    """
    Error general de truora que permite dar mensaje de error como paramtero
    """

    def __init__(self, msg: str):
        super().__init__(msg)

class AutorizacionNoOtorgada(Exception):
    """
    Error por no haber autorizado la
    """

    def __init__(self):
        super().__init__(AUTHORIZATION_EXTERNAL_DB)


"""
Clase pincipal para analysis de creditos financieros

"""
import logging
from typing import Union
from pathlib import Path

from pydantic import BaseModel

from creditpulse.bases_externas.schema import CountryCode, BasesDeDatos, PersonType
from creditpulse.bases_externas.truora import Truora
from creditpulse.bases_externas.database import Database
from creditpulse.common.error_messages import AutorizacionNoOtorgada

logger = logging.getLogger(__name__)


class Check:
    """
        Clase que provee interfaz para analysis the personas
    """

    def __init__(self, truora_api_key: str = None):
        # Module logger

        self.external: Database = Truora(api_key=truora_api_key)
        # TODO implementar otras bases de datos

    def create_check(
            self,
            identificacion: str,
            autorizacion: bool,
            check_type: Union[PersonType, str] = PersonType.COMPANIA,
            check_name: str = 'default',
            force_creation: bool = False

    ):
        """
        Crea un check general en truora
        :return: Analysis financiero de la persona natural o juridica
        """
        if not autorizacion:
            raise AutorizacionNoOtorgada()

        return self.external.create_check(
            identificacion=identificacion,
            person_type=check_type,
            autorizacion_datos=autorizacion,
            pais=CountryCode.COLOMBIA,
            check_name=check_name,
            force_creation=force_creation
        )

    def get_existing_check(self, check_id: str):
        """
        Retorna
        :param check_id:
        :return:
        """
        return self.external.get_check(check_id=check_id)

    def get_existing_check_details(self, check_id: str):
        """

        :param check_id:
        :return:
        """
        return self.external.get_check_details(check_id=check_id)

    def get_existing_check_summary(self, check_id: str):
        """

        :param check_id:
        :return:
        """
        return self.external.get_check_summary(check_id=check_id)

    def to_general(self, check_id: str):
        """
        Create standard output for a check
        :param check_id:
        :return:
        """

        return self.external.to_general(check_id=check_id)

    def get_custom_types(self):
        """

        :return:
        """
        return self.external.configuration.list_configuration()

    def create_default_check(self):
        """

        :return:
        """
        return self.external.configuration.create_default_configuration()

    def delete_custom_check(self, name: str):
        """

        :param name:
        :return:
        """
        model = self.external.configuration.base_model_class.construct(type=name, country=CountryCode.COLOMBIA)
        return self.external.configuration.delete_configuration(model=model)

    def set_up_database(self):
        """

        :return:
        """
        return self.external.configuration.create_default_configuration()

    def create_custom_type(self, model: BaseModel):
        """
        Crea un custom type en truora
        :param model:
        :return:
        """
        return self.external.configuration.create_configuration(model=model)



"""
    Clase abstracta para definir una base de datos
"""
import logging
import requests
from typing import Union
from pydantic import BaseModel, Field, computed_field
from abc import ABC, abstractmethod

from creditpulse.requests.request_manager import RequestActionManager
from creditpulse.bases_externas.schema import PersonType, CountryCode, GeneralDatabase, BasesDeDatos


class DatabaseParams(BaseModel):
    host_url: str = Field(default=None, description="Database host URL")
    version: str = Field(default=None, description="Database api version")
    api_key_name: str = Field(default='DB_API_KEY', description="Database env var name")

    @computed_field
    @property
    def api_url(self) -> str:
        return self.host_url

    @computed_field
    @property
    def base_url(self) -> str:
        return f"{self.api_url}/{self.version}"

    @computed_field
    @property
    def check_url(self) -> str:
        return self.base_url

    @computed_field
    @property
    def check_details_url(self) -> str:
        return self.check_url

    @computed_field
    @property
    def check_summarize_url(self) -> str:
        return self.check_url

    @computed_field
    @property
    def check_settings_url(self) -> str:
        return f"{self.host_url}/{self.version}/settings"

    @computed_field
    @property
    def check_config_url(self) -> str:
        return f"{self.host_url}/{self.version}/config"


class DatabaseConfiguration:
    session: requests.Session
    database_url: str
    base_model_class: BaseModel
    logger = logging.getLogger(__name__)

    def __init__(self, session: requests.Session):
        self.session = session

    @abstractmethod
    def list_configuration(self):
        """
        Return all configuration from database
        :return:
        """

    @abstractmethod
    def update_configuration(self, model: BaseModel):
        """
        Update the given configuration
        :param model:
        :return:
        """

    @abstractmethod
    def create_configuration(self, model: BaseModel):
        """
        Create the given configuration model
        :param model:
        :return:
        """

    @abstractmethod
    def delete_configuration(self, model: BaseModel):
        """
        Delete the given configuration
        :param model:
        :return:
        """

    @abstractmethod
    def create_default_configuration(self):
        """
        Create the default configuration on the database
        :param model:
        :return:
        """

    @abstractmethod
    def reset_default_configuration(self):
        """
        Reset the default configuration
        :return:
        """


class Database(RequestActionManager, ABC):
    configuration: DatabaseConfiguration
    params: DatabaseParams

    @abstractmethod
    def create_check(
            self,
            identificacion: str,
            person_type: Union[PersonType, str],
            autorizacion_datos: bool = False,
            pais: CountryCode = CountryCode.COLOMBIA,
            check_name: str = 'default',
            force_creation: bool = False
    ) -> str:
        """
        Implementacion de
        :param force_creation:
        :param check_name:
        :param identificacion:
        :param person_type:
        :param autorizacion_datos:
        :param pais:
        :return:
        """

    @abstractmethod
    def get_name(self) -> BasesDeDatos:
        """
        Da el nombre de la actual base de datos
        :return:
        """

    @abstractmethod
    def to_general(self, check_id: str) -> GeneralDatabase:
        """
        Traduce base de datos a general
        :return:
        """

    @abstractmethod
    def get_check(self, check_id: str):
        """
        Retorna informacion general del check
        :return:
        """

    @abstractmethod
    def get_check_summary(self, check_id: str):
        """
        retorna informacion resumida del check
        :param check_id:
        :return:
        """

    @abstractmethod
    def get_check_details(self, check_id: str):
        """
        Retorna toda la informacion existente del check
        :param check_id:
        :return:
        """



"""
Clase para hacer exponential back of de algun callback
"""
import logging
import time
from abc import ABC, abstractmethod
import requests

logger = logging.getLogger(__name__)


class RequestActionManager(ABC):
    """Abstract base class for handling HTTP requests with decision logic"""

    @abstractmethod
    def on_execute(self) -> requests.Response:
        """Execute the HTTP request and return the response"""

    @abstractmethod
    def success_callback(self, response: requests.Response) -> None:
        """Handle successful response"""

    @abstractmethod
    def error_callback(self, response: requests.Response) -> None:
        """Handle error response"""

    @abstractmethod
    def is_request_successful(self, response: requests.Response) -> bool:
        """Determine if the request was successful"""


class RequestManager:
    """
        Maneja los request para poder hacer un exponential backoff
    """

    def __init__(self, manager: RequestActionManager, max_retries: int = 12):
        """

        :param manager:
        :param max_retries:
        """

        self.initial_delay = 80
        self.min_delay = 2.0
        self.backoff_factor = 1.5
        self.max_retries = max_retries
        self.manager = manager

        # State variables
        self.current_delay = self.initial_delay
        self.retry_count = 0
        self.running = False

    def reset_delay(self):
        """
        Re-inicia el delay
        :return:
        """
        self.current_delay = self.initial_delay
        self.retry_count = 0

    def increase_delay(self):
        """
        Aumenta el delay cada
        :return:
        """
        self.current_delay = self.initial_delay = max(self.initial_delay / self.backoff_factor, self.min_delay)
        self.retry_count += 1

    def update_backoff_factor(self, amount: float):
        """
        Update back of factor by given amount
        :param amount:
        :return:
        """
        self.backoff_factor = amount
        logger.warning("Backoff updated to {}".format(self.backoff_factor))

    def should_continue(self) -> bool:
        """
        Determina si debe seguir ejecutando la accion
        :return:
        """

        if not self.running:
            return False
        if self.max_retries is not None and self.retry_count >= self.max_retries:
            return False
        return True

    def stop(self):
        """
        Detiene la ejecusion fozosamente
        :return:
        """
        self.running = False
        logger.warning("Completed exp backoff: {}".format(str(id(self))))

    def start(self):
        """
        Inicia el proceso de llamada con backoof
        :return:
        """
        self.running = True
        logger.warning("Staring exp backoff: {}".format(str(id(self))))
        while self.should_continue():
            try:
                response = self.manager.on_execute()

                if self.manager.is_request_successful(response=response):
                    self.stop()
                    self.manager.success_callback(response=response)
                    self.reset_delay()
                else:
                    self.manager.error_callback(response=response)
                    self.increase_delay()

            except requests.exceptions.RequestException:
                self.increase_delay()
            except KeyboardInterrupt:
                self.stop()
            except Exception:
                self.increase_delay()
            if self.should_continue():
                logger.warning('Waiting {} request {}'.format(self.current_delay, str(id(self))))
                time.sleep(self.current_delay)



"""A clean Python project template with public API.

This package provides a simple, clean interface for users to import and use.

Example:
    >>> from CreditPulse import Check
    >>> client = Check()
    >>> result_juridica = client.persona_juridica()
    >>> result_natural = client.persona_natural()
"""

# Import version
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

# Package metadata
__author__ = "Santiago Velez"
__email__ = "s.velez@saman-wm.com"
__description__ = "A python library for financial analysis"

from .check import Check

__all__ = ["Check"]



"""Command line interface."""

import click


@click.command()
@click.option("--name", "-n", help="Name to greet", type=str, default=None, )
@click.version_option()
def main(name: str) -> None:
    """My Project CLI."""
    print(name)



"""
    Request manager ara base de datos externa truora
"""
import json
import os
import logging
import requests
from typing import Optional, Union
from pydantic import BaseModel, Field, computed_field

from creditpulse.requests.request_manager import RequestManager
from creditpulse.bases_externas.schema import (
    TruoraCheckData,
    CountryCode, PersonType,
    CheckStatus,
    GeneralDatabase,
    BasesDeDatos,
    TruoraCustomSchema,
    TruoraCreditScoreCompanySchema,
    TruoraCreditScorePersonSchema,
    parse_json_to_model,
    CreditScoreData,
    PersonalDetails,
    parse_check_to_score,
    Scores,
    PersonaJuridica,
    PersonaNatural
)
from creditpulse.common.error_messages import (
    AutorizacionDatosPersonales,
    TruoraApiKeyRequired,
    TruoraGeneralError
)

from creditpulse.bases_externas.database import Database, DatabaseConfiguration, DatabaseParams

logger = logging.getLogger(__name__)

settings_data = {
    'names_matching_type': 'exact',
    'retries': True,
    'max_duration': '3m'
}


class TruoraParams(DatabaseParams):
    host_url: str = Field(
        default='https://api.checks.truora.com',
        description="Database host URL"
    )

    version: str = Field(
        default='v1',
        description="Database api version"
    )

    @computed_field
    @property
    def api_url(self) -> str:
        return f"{self.host_url}"

    @computed_field
    @property
    def check_url(self) -> str:
        return f"{self.base_url}/checks"

    @computed_field
    @property
    def check_details_url(self) -> str:
        return self.check_url + "/{}/details"

    @computed_field
    @property
    def check_summarize_url(self) -> str:
        return self.check_url + "/checks/{}/summarize"


class TruoraConfiguration(DatabaseConfiguration):
    base_model_class = TruoraCustomSchema
    credit_model_company = TruoraCreditScoreCompanySchema
    credit_model_person = TruoraCreditScorePersonSchema

    def list_configuration(self):
        response = self.session.get(self.database_url)
        if response.status_code not in range(200, 203):
            raise TruoraGeneralError(response.text)
        return json.loads(response.text)

    def update_configuration(self, model: BaseModel):
        return self.session.put(self.database_url, data=model.model_dump())

    def create_configuration(self, model: BaseModel):
        try:
            response = self.session.post(self.database_url, data=model.model_dump())
            if response.status_code not in range(200, 203):
                raise TruoraGeneralError(response.text)
            return json.loads(response.text)
        except Exception as e:
            self.logger.error(msg=str(e))

    def delete_configuration(self, model: BaseModel):
        try:
            params = model.model_dump(include={'type', 'country'})
            response = self.session.delete(self.database_url, params=params)
            if response.status_code not in range(200, 203):
                raise TruoraGeneralError(response.text)
            return json.loads(response.text)
        except Exception as e:
            self.logger.error(msg=str(e))

    def reset_default_configuration(self):
        return self.update_configuration(model=self.base_model_class())

    def create_default_configuration(self):

        models = [self.base_model_class(), self.credit_model_company(), self.credit_model_person()]
        for model in models:
            try:
                update_response = self.update_configuration(model=model)
                self.logger.info('Creating default model')
                if update_response.status_code in range(200, 204):
                    self.logger.info(msg="Database has been updated: default model")
                elif update_response.status_code == 400 and update_response.json()['code'] == 10400:
                    self.logger.error(msg="Database configuration does not exists")
                    try:
                        response = self.create_configuration(model=model)
                        self.logger.info(msg=response)
                    except Exception as e:
                        self.logger.error(msg=str(e))

            except Exception as e:
                self.logger.error(msg=str(e))


class Truora(Database):
    """

    Clase principal para consultar base de datos externa truora
    """

    def __init__(self, api_key: str = None):
        """

        :param api_key:
        """

        self.params = TruoraParams()

        self.api_key = api_key if api_key is not None else os.environ.get(self.params.api_key_name)

        if self.api_key is None:
            e = TruoraApiKeyRequired()
            logger.error(e)
            raise e

        self.session = requests.session()

        self.session.headers.update({
            "Truora-API-Key": self.api_key,
            'Accept': 'application/json'
        })

        self.request_manager = RequestManager(
            manager=self,
            max_retries=30
        )

        self.tcheck: Optional[TruoraCheckData] = None

        self.status: CheckStatus = CheckStatus.NOT_STARTED

        self.configuration = TruoraConfiguration(session=self.session)
        self.configuration.database_url = self.params.check_config_url

    def get_name(self) -> BasesDeDatos:
        return BasesDeDatos.TRUORA

    def _get_check(self, check_id: str):
        return self.session.get(self.params.check_url + "/" + check_id)

    def on_execute(self) -> requests.Response:
        return self._get_check(self.tcheck.check.check_id)

    def success_callback(self, response: requests.Response) -> None:
        logger.info("Consulta a Base De datos ha sido finalizada")
        response_json = response.json()
        self.tcheck = TruoraCheckData(**response_json)
        self.status = self.tcheck.check.status

    def error_callback(self, response: requests.Response) -> None:
        response_json = response.json()
        inter_check: TruoraCheckData = TruoraCheckData(**response_json)

        if inter_check.check.status == CheckStatus.DELAYED:
            self.request_manager.update_backoff_factor(self.request_manager.backoff_factor + 0.5)
            logger.warning("Consulta base de datos ha sido delayed")

        self.status = inter_check.check.status

    def is_request_successful(self, response: requests.Response) -> bool:
        """

        :param response:
        :return:
        """
        if response.status_code in [200, 203]:
            response_json = response.json()
            try:
                t_checker: TruoraCheckData = TruoraCheckData(**response_json)
                return t_checker.check.status == CheckStatus.COMPLETED
            except:
                return False
        return False

    def create_check(
            self,
            identificacion: str,
            person_type: Union[PersonType, str],
            autorizacion_datos: bool = False,
            pais: CountryCode = CountryCode.COLOMBIA,
            check_name: str = 'default',
            force_creation: bool = False
    ) -> str:
        """
        Funcion principal para consultar base de datos externa truora

        :param identificacion:
        :param person_type:
        :param autorizacion_datos:
        :param pais:
        :param check_name:
        :param force_creation:
        :return:
        """

        if not autorizacion_datos:
            raise AutorizacionDatosPersonales()

        id_field_name = 'national_id'
        if person_type == PersonType.COMPANIA:
            id_field_name = 'tax_id'

        form_data = {
            id_field_name: identificacion,
            'country': pais,
            'type': check_name,
            'user_authorized': autorizacion_datos,
            'force_creation': force_creation
        }
        response = self.session.post(self.params.check_url, data=form_data)
        response_json = response.json()

        if response.status_code not in [200, 201, 202, 203]:
            logger.error(f"Error al crear consulta base de datos externa: {response_json['message']}")
            raise TruoraGeneralError(f"Error al crear consulta base de datos externa: {response_json['message']}")

        self.tcheck = TruoraCheckData(**response_json)

        if self.tcheck is None:
            raise TruoraGeneralError('Check de truora no fue creado')

        self.request_manager.start()

        return self.tcheck.check.check_id

    def get_check(self, check_id: str):
        """

        :param check_id:
        :return:
        """
        response = self._get_check(check_id=check_id)
        if response.status_code not in range(200, 203):
            logger.error(response.text)
            raise TruoraGeneralError(response.text)

        return json.loads(response.text)

    def get_check_summary(self, check_id: str):
        """

        :param check_id:
        :return:
        """
        response = self.session.get(self.params.check_summarize_url.format(check_id))
        if response.status_code not in range(200, 203):
            logger.error(response.text)
            raise TruoraGeneralError(response.text)

        return json.loads(response.text)

    def get_check_details(self, check_id: str):
        """

        :param check_id:
        :return:
        """
        response = self.session.get(self.params.check_details_url.format(check_id))
        if response.status_code not in range(200, 203):
            logger.error(response.text)
            raise TruoraGeneralError(response.text)

        return json.loads(response.text)

    def get_next_details(self, next_url: str):
        """

        :param next_url:
        :return:
        """
        response = self.session.get("{}/{}".format(self.params.api_url, next_url))
        if response.status_code not in range(200, 203):
            logger.error(response.text)
            raise TruoraGeneralError(response.text)

        return json.loads(response.text)

    def to_general(self, check_id: str) -> GeneralDatabase:
        """
        Traduce truora a general
        :return:
        """
        check_data = self.get_check(check_id=check_id)
        scores = parse_check_to_score(json_data=check_data, model_class=Scores)

        details = self.get_check_details(check_id=check_id)
        credit_data = parse_json_to_model(json_data=details, model_class=CreditScoreData)

        personal_details = PersonalDetails()

        next = details.get('next')

        juridica = parse_json_to_model(json_data=details, model_class=PersonaJuridica)

        natural = parse_json_to_model(json_data=details, model_class=PersonaNatural)

        if next or not isinstance(next, str):
            next_details = self.get_next_details(next_url=details.get('next'))
            personal_details = parse_json_to_model(json_data=next_details, model_class=PersonalDetails)

        return GeneralDatabase(
            check_id=check_id,
            credit_data=credit_data,
            personal_data=personal_details,
            scores=scores,
            juridica=juridica,
            natural=natural
        )



import unittest

from creditpulse.bases_externas.truora import Truora
from creditpulse.bases_externas.schema import CountryCode, PersonType, TruoraCustomSchema


class TestTruoraCheck(unittest.TestCase):

    def test_simple_request_persona(self):
        client = Truora()
        client.create_check(
            identificacion='1053778047',
            person_type=PersonType.PERSONA,
            pais=CountryCode.COLOMBIA,
            autorizacion_datos=True
        )

    def test_simple_request_compania(self):
        client = Truora()
        client.create_check(
            identificacion='1053778047',
            person_type=PersonType.COMPANIA,
            pais=CountryCode.COLOMBIA,
            autorizacion_datos=True
        )

    def test_custom_type(self):
        client = Truora()
        check_name = "StandardCheck"
        schema = TruoraCustomSchema(
            type=check_name,
            country="CO",
        )
        response = client.create_custom_type(model=schema)
        print(response)

        client.create_check(
            identificacion='1053778047',
            person_type=check_name,
            pais=CountryCode.COLOMBIA,
            autorizacion_datos=True
        )



"""
Definicion de tipos de datos para base de datos externas
"""
import re
from datetime import datetime
from enum import Enum
from typing import Dict, Union, Optional, Any, Type, TypeVar
from pydantic import BaseModel, Field


class BasesDeDatos(str, Enum):
    """
    
    Bases de Datos actualmente soportadas
    """
    TRUORA = "truora"


class CountryCode(str, Enum):
    """Supported country codes for checks"""
    ALL = "ALL"  # International Lists
    BRAZIL = "BR"  # Brazil
    COLOMBIA = "CO"  # Colombia
    CHILE = "CL"  # Chile
    COSTA_RICA = "CR"  # Costa Rica
    MEXICO = "MX"  # Mexico
    PERU = "PE"  # Peru


class PersonType(str, Enum):
    """Supported country codes for checks"""
    PERSONA = "person"  # Used to perform a background check on a person.
    VEHICULO = "vehicle"  # Used to perform a background check on a driver and their vehicle
    COMPANIA = "company"  # Used to perform a background check on a company


class CheckStatus(str, Enum):
    """Status values for checks"""
    NOT_STARTED = "not_started"  # The check is enqueued and the data collection has not started yet
    IN_PROGRESS = "in_progress"  # Data is being collected but some data sources may have finished already
    DELAYED = "delayed"  # One or more data sources are taking a long time to query the data. Most data sources will have already finished
    COMPLETED = "completed"  # The check finished and 70% or more of the data sources did not end in error status
    ERROR = "error"  # The check finished and more than 30% of the data sources ended in error status


class TruoraCheck(BaseModel):
    """
    Definicion para Check de truora
    """
    check_id: str = Field(..., description="Unique identifier for the check")
    country: CountryCode = Field(
        ...,
        description="""
            Country code - 
            ALL for International Lists, 
            BR for Brazil, 
            CO for Colombia, 
            CL for Chile, 
            CR for Costa Rica, 
            MX for Mexico, 
            PE for Peru
        """
    )
    type: str = Field(..., description="Check type")
    status: CheckStatus = Field(..., description="Current status of the check")
    creation_date: datetime = Field(..., description="When the check was created")
    update_date: datetime = Field(..., description="When the check was last updated")
    national_id: str = Field(default=None, description="National ID number being checked")
    tax_id: str = Field(default=None, description="National ID number being checked")

    # These are optional
    name_score: Optional[float] = Field(default=0, description="Score for name matching")
    id_score: Optional[float] = Field(default=-1, description="Score for ID matching")
    score: Optional[float] = Field(default=-1, description="Overall check score")
    previous_check: Optional[str] = Field(default=None, description="Overall check score")


class TruoraCheckData(BaseModel):
    """
    Definicion para la respuesta de truora
    """
    check: TruoraCheck = Field(..., description="Check details")
    details: Optional[str] = Field(default=None, description="API endpoint for detailed check information")
    self: Optional[str] = Field(default=None, description="API endpoint for this check resource")


class TruoraCustomSchema(BaseModel):
    type: str = Field(default='default', description="CustomTypeName")
    country: CountryCode = Field(
        default=CountryCode.COLOMBIA,
        description="""
            Country code - 
            ALL for International Lists, 
            BR for Brazil, 
            CO for Colombia, 
            CL for Chile, 
            CR for Costa Rica, 
            MX for Mexico, 
            PE for Peru
        """
    )

    dataset_affiliations_and_insurances: float = Field(default=0.1, description="dataset_affiliations_and_insurances")
    dataset_taxes_and_finances: float = Field(default=0.1, description="dataset_taxes_and_finances")
    dataset_legal_background: float = Field(default=0.1, description="dataset_legal_background")
    dataset_criminal_record: float = Field(default=0.1, description="dataset_criminal_record")
    dataset_personal_identity: float = Field(default=0.1, description="dataset_personal_identity")
    dataset_professional_background: float = Field(default=0.1, description="dataset_professional_background")
    dataset_international_background: float = Field(default=0.1, description="dataset_international_background")
    dataset_business_background: float = Field(default=0.1, description="dataset_business_background")
    dataset_credit_history: float = Field(default=0.1, description="dataset_credit_history")
    dataset_document_validation: float = Field(default=0.05, description="dataset_document_validation")
    dataset_alert_in_media: float = Field(default=0.05, description="dataset_alert_in_media")

    # dataset_traffic_fines: float = Field(default=0, description="dataset_traffic_fines")
    # dataset_driving_licenses: float = Field(default=0, description="dataset_driving_licenses")
    # dataset_vehicle_information: float = Field(default=0, description="dataset_vehicle_information")
    # dataset_vehicle_permits: float = Field(default=0, description="dataset_vehicle_permits")


class TruoraCreditScoreCompanySchema(BaseModel):
    type: str = Field(default='credit-company', description="CustomTypeName")
    country: CountryCode = Field(
        default=CountryCode.COLOMBIA,
        description="""
            Country code - 
            ALL for International Lists, 
            BR for Brazil, 
            CO for Colombia, 
            CL for Chile, 
            CR for Costa Rica, 
            MX for Mexico, 
            PE for Peru
        """
    )

    dataset_credit_history: float = Field(default=0.4, description="dataset_credit_history")#8
    dataset_business_background: float = Field(default=0.1, description="dataset_business_background")#7
    dataset_legal_background: float = Field(default=0.1, description="dataset_legal_background")#6
    dataset_affiliations_and_insurances: float = Field(default=0.1, description="dataset_affiliations_and_insurances")#5
    dataset_taxes_and_finances: float = Field(default=0.1, description="dataset_taxes_and_finances")#4
    dataset_alert_in_media: float = Field(default=0.05, description="dataset_alert_in_media")#3
    dataset_international_background: float = Field(default=0.1, description="dataset_international_background")#2
    dataset_criminal_record: float = Field(default=0.05, description="dataset_criminal_record")#1

class TruoraCreditScorePersonSchema(BaseModel):
    type: str = Field(default='credit-person', description="CustomTypeName")
    country: CountryCode = Field(
        default=CountryCode.COLOMBIA,
        description="""
            Country code - 
            ALL for International Lists, 
            BR for Brazil, 
            CO for Colombia, 
            CL for Chile, 
            CR for Costa Rica, 
            MX for Mexico, 
            PE for Peru
        """
    )

    dataset_credit_history: float = Field(default=0.4, description="dataset_credit_history")#1
    dataset_personal_identity: float = Field(default=0.1, description="dataset_personal_identity")#2
    dataset_legal_background: float = Field(default=0.1, description="dataset_legal_background")#3
    dataset_affiliations_and_insurances: float = Field(default=0.1, description="dataset_affiliations_and_insurances")#4
    dataset_taxes_and_finances: float = Field(default=0.1, description="dataset_taxes_and_finances")#5
    dataset_alert_in_media: float = Field(default=0.05, description="dataset_alert_in_media")#6
    dataset_international_background: float = Field(default=0.1, description="dataset_international_background")#7
    dataset_criminal_record: float = Field(default=0.05, description="dataset_criminal_record")#8

class PersonaNatural(BaseModel):
    estado_afiliacion: Optional[str] = Field(
        default=None,
        description="Estado",
        title="Información afiliación"
    )
    tipo_afiliacion: Optional[str] = Field(
        default=None,
        description="Tipo de afiliado",
        title="Información afiliación"
    )
    regimen_afiliacion: Optional[str] = Field(
        default=None,
        description="Regimen",
        title="Información afiliación"
    )


class PersonaJuridica(BaseModel):
    razon_social: Optional[str] = Field(
        default=None,
        description="Razón social",
        title="Información General"
    )

    estado: Optional[str] = Field(
        default=None,
        description="Estado de Matrícula",
        title="Información General"
    )

    fecha_matricula: Optional[str] = Field(
        default=None,
        description="Fecha de Matrícula",
        title="Registro Mercantil"
    )

    representante_legal: Optional[str] = Field(
        default=None,
        description="Nombre",
        title="Representantes legales"
    )

    documento_representante_legal: Union[str, int, float, None] = Field(
        default=None,
        description="Número de documento",
        title="Representantes legales"
    )


class Scores(BaseModel):
    general: Optional[float] = Field(
        default=None,
        description="Overal Score",
    )

    credit_history: Optional[float] = Field(
        default=None,
        description="credit_history",
        title="credit_history",
    )

    criminal_record: Optional[float] = Field(
        default=None,
        description="criminal_record",
        title="criminal_record",
    )

    legal_background: Optional[float] = Field(
        default=None,
        description="legal_background",
        title="legal_background",
    )

    international_background: Optional[float] = Field(
        default=None,
        description="international_background",
        title="international_background",
    )

    personal_identity: Optional[float] = Field(
        default=None,
        description="personal_identity",
        title="personal_identity",
    )

    business_background: Optional[float] = Field(
        default=None,
        description="business_background",
        title="business_background",
    )

    alert_in_media: Optional[float] = Field(
        default=None,
        description="alert_in_media",
        title="alert_in_media",
    )

    taxes_and_finances: Optional[float] = Field(
        default=None,
        description="taxes_and_finances",
        title="taxes_and_finances",
    )


class PersonalDetails(BaseModel):
    nombre: Optional[str] = Field(
        default=None,
        description="Nombre Completo",
        title="Información general"
    )

    nuip: Union[str, int, None] = Field(
        default=None,
        description="NUIP",
        title="Información general"
    )


class CreditScoreData(BaseModel):
    """Modelo para datos extraídos del cliente"""

    score: int = Field(
        default=0,
        ge=0,
        le=1000,
        description="Score",
        title="Score"
    )

    mora_actual: float = Field(
        default=0,
        ge=0,
        description="Días de mora actual",
        title="Mora Maxima del portafolio a la fecha"
    )

    mora_maxima_6m: float = Field(
        default=0, ge=0,
        description="Mora máxima últimos 6 meses",
        title="Mora Maxima en los Ultimos 6 meses"
    )

    mora_maxima_12m: float = Field(
        default=0, ge=0,
        description="Mora máxima últimos 12 meses",
        title="Mora Maxima en los Ultimos 12 meses"
    )

    mora_historica: float = Field(
        default=0, ge=0,
        description="Mora Maxima del portafolio a la fecha",
        title="Desempeño actual en todos los portafolios"
    )

    productos_al_dia: float = Field(
        default=0, ge=0,
        description="Numero total de productos al dia a la fecha",
        title="Desempeño actual en todos los portafolios"
    )

    productos_castigados: float = Field(
        default=0, ge=0,
        description="Numero total de productos castigados a la fecha",
        title="Desempeño actual en todos los portafolios"
    )

    ever_120_6m: float = Field(
        default=0, ge=0,
        description="productos Ever 120 en los ultimos 6 meses",
        title="Desempeño a corto plazo en todos los portafolios"
    )

    ever_90_6m: float = Field(
        default=0, ge=0,
        description="productos Ever 90 en los ultimos 6 meses",
        title="Desempeño a corto plazo en todos los portafolios"
    )

    ever_60_6m: float = Field(
        default=0, ge=0,
        description="productos Ever 60 en los ultimos 6 meses",
        title="Desempeño a corto plazo en todos los portafolios"
    )

    ever_30_6m: float = Field(
        default=0, ge=0,
        description="productos Ever 30 en los ultimos 6 meses",
        title="Desempeño a corto plazo en todos los portafolios"
    )

    ever_120_12m: float = Field(
        default=0, ge=0,
        description="productos Ever 120 en los ultimos 12 meses",
        title="Desempeño a mediano plazo en todos los portafolios"
    )

    ever_90_12m: float = Field(
        default=0, ge=0,
        description="productos Ever 90 en los ultimos 12 meses",
        title="Desempeño a mediano plazo en todos los portafolios"
    )

    ever_60_12m: float = Field(
        default=0, ge=0,
        description="productos Ever 60 en los ultimos 12 meses",
        title="Desempeño a mediano plazo en todos los portafolios"
    )

    ever_30_12m: float = Field(
        default=0, ge=0,
        description="productos Ever 30 en los ultimos 12 meses",
        title="Desempeño a mediano plazo en todos los portafolios"
    )

    siempre_al_dia_6m: float = Field(
        default=0, ge=0,
        description="Productos siempre al día 6 meses",
        title="Productos siempre al día 6 meses"
    )

    siempre_al_dia_12m: int = Field(
        default=0, ge=0,
        description="Productos siempre al día 12 meses",
        title="Productos siempre al día 12 meses"
    )

    cuota_total: float = Field(
        default=0.0, ge=0,
        description="Cuota total en salarios mínimos",
        title="Cuota total en salarios mínimos"
    )

    saldo_rotativo: float = Field(
        default=0.0, ge=0,
        description="Saldo rotativo",
        title="Saldo rotativo"
    )

    saldo_consumo: float = Field(
        default=0.0, ge=0,
        description="Saldo consumo",
        title="Saldo consumo"
    )

    estrato: float = Field(
        default=0.0, ge=0, le=8,
        description="Estrato socioeconómico",
        title="Estrato socioeconómico"
    )

    estado_civil: str = Field(
        default="soltero",
        description="Estado civil",
        title="Estado civil"
    )

    nivel_estudios: str = Field(
        default="primaria",
        description="Nivel de estudios",
        title="Nivel de estudios"
    )

    # Campos calculados
    saldo_total: float = Field(
        default=0.0, ge=0,
        description="Saldo total calculado"
    )

    productos_totales: float = Field(
        default=0.0, ge=0,
        description="Total productos"
    )

    # experiencia financiera

    meses_producto_antiguo: float = Field(
        default=0.0, ge=0,
        description="Meses desde la apertura del producto mas antiguo en todos los portafolios",
        title="Experiencia financiera"
    )

    meses_producto_abierto_antiguo: float = Field(
        default=0.0, ge=0,
        description="Meses desde la apertura del producto abierto mas antiguo en todos los portafolios",
        title="Experiencia financiera"
    )
    meses_producto_reciente: float = Field(
        default=0.0, ge=0,
        description="Meses desde la apertura del producto mas reciente en todos los portafolios",
        title="Experiencia financiera"
    )

    numero_productos_creditos_rotativos: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio: créditos rotativos, cartera bancaria, microcrédito, cartera compañías de financiamiento comercial, Tarjetas de Crédito",
        title="Experiencia financiera"
    )

    numero_productos_tarjeta_credito: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio tarjetas de crédito",
        title="Experiencia financiera"
    )

    numero_productos_fondo_empleados: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio: cartera bancaria, cartera de corporaciones financieras, cartera de fondos de empleados, cartera de compañías de leasing, cartera compañías de financiamiento comercial, cartera de compensación y salud, créditos de consumo, microcrédito",
        title="Experiencia financiera"
    )

    numero_productos_credito_automotriz: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio: cartera automotriz",
        title="Experiencia financiera"
    )

    numero_productos_credito_vivienda: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio: cartera de vivienda",
        title="Experiencia financiera"
    )

    numero_productos_cuenta_corriente: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio cuenta corriente",
        title="Experiencia financiera"
    )

    numero_productos_otros_tipos: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio otros tipos de cartera",
        title="Experiencia financiera"
    )

    numero_productos_cuenta_ahorros: float = Field(
        default=0.0, ge=0,
        description="Numero de Productos Abiertos en el portafolio cuentas de ahorros bancarias",
        title="Experiencia financiera"
    )

    numero_productos_codeudores: float = Field(
        default=0.0,
        description="Numero de Productos Abiertos en el portafolio codeudores",
        title="Experiencia financiera"
    )

    cuota_todos_portafolio: float = Field(
        default=0.0,
        description="Cuota total en productos a la fecha en todos los portafolios",
        title="Endeudamiento financiero"
    )

    saldo_total_rotativos: float = Field(
        default=0.0,
        description="Saldo total en productos a la fecha en el portafolio cartera bancaria, cartera de corporaciones financieras, cartera de fondos de empleados, cartera de compañías de leasing, cartera compañías de financiamiento comercial, cartera de compensación y salud, créditos de consumo, microcrédito",
        title="Endeudamiento financiero"
    )


def extract_numeric_value_or_leave(text: str):
    """
    Extrae valor numérico de un texto si empieza con número, sino retorna el texto original
    :param text: Textio a extraer datos numericos
    :return:
    """
    if not text or text in ["No tiene información en el portafolio", "No tiene cuentas abiertas", None]:
        return 0.0

    text = str(text).strip()

    tokens = text.split(" ")

    if len(tokens) > 1:
        match = re.match(r'^(\d+(?:\.\d+)?)', tokens[0])
        if match:
            return float(match.group(1))
        else:
            return text
    else:
        return text


T = TypeVar('T', bound=BaseModel)


def _is_valid_dict(obj: Any) -> bool:
    """
    Verifica si el objeto es un diccionario válido y no vacío.
    :param obj:
    :return:
    """
    return obj is not None and isinstance(obj, dict)


def _get_field_title_safely(field_info: Any, field_name: str) -> str:
    """
    Obtiene el título del campo de forma segura.
    :param field_info:
    :return:
    """
    try:
        # Intentar obtener title como atributo
        if hasattr(field_info, field_name) and getattr(field_info, field_name):
            return str(getattr(field_info, field_name)).lower()

        if isinstance(field_info, dict) and field_name in field_info:
            title = field_info.get(field_name)
            if title:
                return str(title).lower()

        return ""

    except Exception:
        return ""


def _safe_string_extract(value: Any) -> str:
    """
    Extrae string de forma segura, manejando None y otros tipos.
    :param value:
    :return:
    """
    if value is None:
        return ""

    try:
        return str(value).strip()
    except Exception:
        return ""


def _process_cell_safely(cell: Any, raw_data: Dict[str, Any], model_class: Type[T], section_title: str) -> None:
    """
    Procesa una celda individual de forma segura.

    :param cell: Celda a procesar
    :param raw_data: Diccionario donde almacenar los datos extraídos
    :param model_class: Clase del modelo para obtener los campos
    """
    if not _is_valid_dict(cell):
        return

    # Extraer label y value de forma segura
    label = _safe_string_extract(cell.get('label'))
    value = _safe_string_extract(cell.get('value'))

    # Skip si no hay label válido
    if not label:
        return

    # Buscar coincidencias con los campos del modelo
    label_lower = label.lower()

    try:
        for field_name, field_info in model_class.model_fields.items():

            # Skip si ya tenemos este campo
            if field_name in raw_data:
                continue

            # Obtener el título del campo de forma segura
            field_description = _get_field_title_safely(field_info=field_info, field_name='description')
            field_title = _get_field_title_safely(field_info=field_info, field_name='title')

            if len(field_description) > 0:
                if section_title.lower() in field_title.lower() or field_title.lower() in section_title.lower():
                    # Verificar coincidencia
                    if field_description and field_description in label_lower:
                        try:
                            extracted_value = extract_numeric_value_or_leave(value)
                            raw_data[field_name] = extracted_value

                            break  # Salir del loop una vez encontrada la coincidencia

                        except Exception as e:

                            continue

    except Exception as e:
        pass


def _filter_valid_data(raw_data: Dict[str, Any], model_class: Type[T]) -> Dict[str, Any]:
    """
    Filtra los datos para mantener solo los que son válidos para el modelo.

    :param raw_data: Datos sin filtrar
    :param model_class: Clase del modelo
    :return: Datos filtrados
    """
    valid_data = {}

    try:
        model_fields = getattr(model_class, 'model_fields', {})

        for field_name, value in raw_data.items():
            if field_name in model_fields:
                # Intentar validar el campo individualmente
                try:
                    # Crear un modelo temporal solo con este campo
                    temp_data = {field_name: value}
                    model_class(**temp_data)
                    valid_data[field_name] = value
                except Exception:
                    continue

    except Exception as e:
        pass

    return valid_data


def _create_model_safely(model_class: Type[T], raw_data: Dict[str, Any]) -> T:
    """
    Crea el modelo de forma segura con manejo de errores.

    :param model_class: Clase del modelo a crear
    :param raw_data: Datos extraídos
    :return: Instancia del modelo
    """
    try:
        if raw_data:
            return model_class(**raw_data)
        else:
            return model_class()

    except Exception as e:

        # Intentar crear con un subconjunto de datos válidos
        try:
            valid_data = _filter_valid_data(raw_data, model_class)
            if valid_data:
                return model_class(**valid_data)
        except Exception as e2:
            pass

        # Retornar con valores por defecto si todo falla
        return model_class()


def parse_json_to_model(json_data: Dict[str, Any], model_class: Type[T]) -> T:
    """
    Extrae y valida los datos del JSON usando Pydantic de forma segura.

    :param json_data: Data 'cruda' para extraer los datos de
    :param model_class: Clase del modelo Pydantic a crear
    :return: Instancia del modelo con datos extraídos o valores por defecto
    """
    raw_data = {}

    try:
        # Validar entrada básica
        if not isinstance(json_data, dict):
            return model_class()

        if not hasattr(model_class, 'model_fields'):
            return model_class()

        # Obtener detalles de forma segura
        details = json_data.get('details')
        if not details or not isinstance(details, list):
            return model_class()

        # Procesar cada detalle
        for detail_idx, detail in enumerate(details):
            if not _is_valid_dict(detail):
                continue

            # Obtener tablas de forma segura
            tables = detail.get('tables')
            if not tables or not isinstance(tables, list):
                continue

            # Procesar cada tabla
            for table_idx, table in enumerate(tables):
                if not _is_valid_dict(table):
                    continue

                # Obtener filas de forma segura
                rows = table.get('rows')
                if not rows or not isinstance(rows, list):
                    continue

                section_title = table.get('title')

                if not section_title or not isinstance(section_title, str):
                    continue

                # Procesar cada fila
                for row_idx, row in enumerate(rows):
                    if not _is_valid_dict(row):
                        continue

                    # Obtener celdas de forma segura
                    cells = row.get('cells')
                    if not cells or not isinstance(cells, list):
                        continue

                    # Procesar cada celda
                    for cell_idx, cell in enumerate(cells):
                        try:
                            _process_cell_safely(
                                cell=cell,
                                raw_data=raw_data,
                                model_class=model_class,
                                section_title=section_title
                            )
                        except Exception as e:
                            continue

    except Exception as e:
        return model_class()

    # Crear y validar el modelo Pydantic
    return _create_model_safely(model_class, raw_data)


def parse_check_to_score(json_data: Dict[str, Any], model_class: Type[T]) -> T:
    """
    Extrae y valida los datos del JSON usando Pydantic de forma segura.

    :param json_data: Data 'cruda' para extraer los datos de
    :param model_class: Clase del modelo Pydantic a crear
    :return: Instancia del modelo con datos extraídos o valores por defecto
    """
    raw_data = {}

    try:
        # Validar entrada básica
        if not isinstance(json_data, dict):
            return model_class()

        if not hasattr(model_class, 'model_fields'):
            return model_class()

        check = json_data.get('check')

        if not check or not isinstance(check, dict):
            return model_class()

        raw_data['general'] = check.get('score', 0.0)
        scores = check.get('scores')

        if not scores or not isinstance(scores, list):
            return model_class()

        for cell_idx, score in enumerate(scores):
            if "data_set" in score and "score" in score:
                raw_data[score["data_set"]] = score["score"]

        return _create_model_safely(model_class, raw_data)
    except Exception as e:
        return model_class()


class GeneralDatabase(BaseModel):
    check_id: str = Field(..., description="Check details")
    credit_data: CreditScoreData = Field(..., description="Credit Score Details")
    personal_data: PersonalDetails = Field(..., description="Personal Data")
    scores: Scores = Field(..., description="Dataset Scores")
    juridica: PersonaJuridica = Field(..., description="Datos persona Juridica")
    natural: PersonaNatural = Field(..., description="Datos persona Natural")



from creditpulse.requests.request_manager import RequestManager, RequestActionManager

import unittest
import requests


class SimpleCheck(RequestActionManager):

    def __init__(self, endpoint):
        self.endpoint = endpoint
        self.success_called = False
        self.error_called = False

    def on_execute(self) -> requests.Response:
        return requests.get(self.endpoint)

    def success_callback(self, response: requests.Response) -> None:
        self.success_called = True

    def error_callback(self, response: requests.Response) -> None:
        self.error_called = True

    def is_request_successful(self, response: requests.Response) -> bool:
        return response.status_code in [200, 201]


class TestRequestManager(unittest.TestCase):

    def test_request_success(self):
        check = SimpleCheck(endpoint='https://httpbin.org/get')
        manager = RequestManager(
            manager=check,
            max_retries=1
        )

        manager.start()

        self.assertTrue(check.success_called, "Success callback was not called")

    def test_request_error(self):
        self.error_callback_code = 200
        self.retries = 3
        check = SimpleCheck(endpoint='https://httpbin.org/status/500')

        manager = RequestManager(
            manager=check,
            max_retries=self.retries,
        )

        manager.start()
        self.assertTrue(check.error_called, "Success callback was not called")
        self.assertEqual(manager.retry_count, self.retries)


if __name__ == '__main__':
    unittest.main()
